/** @file mission_node.h
 *  @version 4.0
 *  @date June 2020
 *
 *  @brief node of hotpoint 1.0/waypoint 1.0.
 *
 *  @Copyright (c) 2020 DJI
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#ifndef MISSION_NODE_H
#define MISSION_NODE_H

// System includes
#include "unistd.h"
#include <iostream>
#include <string>

// DJI SDK includes
#include <dji_osdk_ros/MissionWpAction.h>
#include <dji_osdk_ros/MissionWpUpload.h>
#include <dji_osdk_ros/Overwatch.h>
#include <dji_osdk_ros/vehicle_wrapper.h>
#include <dji_osdk_ros/dji_vehicle_node.h>

// ROS includes
#include <ros/ros.h>
#include <sensor_msgs/NavSatFix.h>
//#include <dji_vehicle.hpp>


#define C_EARTH (double)6378137.0
#define C_PI (double)3.141592653589793
#define DEG2RAD(DEG) ((DEG) * ((C_PI) / (180.0)))

typedef struct ServiceAck
{
  bool         result;
  int          cmd_set;
  int          cmd_id;
  unsigned int ack_data;
  ServiceAck(bool res, int set, int id, unsigned int ack)
    : result(res)
    , cmd_set(set)
    , cmd_id(id)
    , ack_data(ack)
  {
  }
  ServiceAck()
  {
  }
} ServiceAck;


float eucDistance(std::vector<float> a, std::vector<float> b);
float normalizeAngle(float angle);
std::vector<std::vector<float>> rotateAxisAngle(std::vector<float> u, float angleRad);
float dotProduct(std::vector<float> vect_A, std::vector<float> vect_B);
std::vector<float> crossProduct(std::vector<float> vect_A, std::vector<float> vect_B);
std::vector<float> normalizeVector(std::vector<float> vec);
float Magnitude3(std::vector<float> v);
std::vector<float> MatrixMultiply(std::vector <std::vector<float>> a, std::vector<float> b);
std::vector<std::vector<float>> GetRotationMatrixParameter();

bool setupWaypointMission(int responseTimeout);
bool runWaypointMission();
bool runPositionMission();

void setWaypointDefaults(DJI::OSDK::WayPointSettings* wp);

void setWaypointInitDefaults(dji_osdk_ros::MissionWaypointTask& waypointTask);

// std::vector<DJI::OSDK::WayPointSettings> createWaypoints();

std::vector<DJI::OSDK::WayPointSettings> generateWaypointsPolygon(
  DJI::OSDK::WayPointSettings* start_data, DJI::OSDK::float64_t increment,
  int num_wp);

std::vector<DJI::OSDK::WayPointSettings> loadWaypoints();

void uploadWaypoints(std::vector<DJI::OSDK::WayPointSettings>& wp_list,
                     int                                       responseTimeout,
                     dji_osdk_ros::MissionWaypointTask&        waypointTask);

ServiceAck initWaypointMission(dji_osdk_ros::MissionWaypointTask& waypointTask);

ServiceAck missionAction(DJI::OSDK::DJI_MISSION_TYPE type,
                         DJI::OSDK::MISSION_ACTION   action);

ServiceAck activate();

ServiceAck obtainCtrlAuthority();

bool takeoff();

void relativeVelocity(float x, float y, float z, float yaw, float dt);
void relativePosition(float x, float y, float z, float yaw, float dt);

float applyRotationAngle(float targetYaw);

bool land();

bool setRates();

bool stopMotion();

bool goHome();

void calibratePositionReference(std::vector<float> requestedPosition, std::vector<float> actualPosition);

void updateMagnitudeScaler(float newValue);

float flatAngleBetween(std::vector<float> a, std::vector<float> b);

void updateAngleScaler(float newValue);

bool overwatchFunction(dji_osdk_ros::Overwatch::Request &req, dji_osdk_ros::Overwatch::Response &res);
// std::vector<float> DoCalibrationMove(std::vector<float> calibrationVec);

void gpsPosCallback(const sensor_msgs::NavSatFix::ConstPtr& msg);
void posFeedbackCallback(const geometry_msgs::Transform::ConstPtr& msg);
void pdPositionCallback(const geometry_msgs::Transform::ConstPtr& msg);
void posTrackingCallback(const geometry_msgs::Transform::ConstPtr& msg);

void pdMaxSpeedCallback(const std_msgs::Float32::ConstPtr& msg);
void pdMaxYawSpeedCallback(const std_msgs::Float32::ConstPtr& msg);
void pdSpeedScaleCallback(const std_msgs::Float32::ConstPtr& msg);
void pdUpdateRateCallback(const std_msgs::Float32::ConstPtr& msg);
void pdDampingRateCallback(const std_msgs::Float32::ConstPtr& msg);
void pdFrequencyRateCallback(const std_msgs::Float32::ConstPtr& msg);
void pdAngleOffsetRateCallback(const std_msgs::Float32::ConstPtr& msg);



#endif // MISSION_NODE_H
