#!/usr/bin/python3

import rospy
import numpy as np
from sensor_msgs.msg import NavSatFix
from sensor_msgs.msg import Imu
from std_msgs.msg import Float32
from geometry_msgs.msg import PointStamped
from geometry_msgs.msg import QuaternionStamped
from geometry_msgs.msg import Transform
from geometry_msgs.msg import Vector3
from geometry_msgs.msg import Quaternion
import math
from pyproj import Transformer

class QualisysSpoof():
    def __init__(self):
        rospy.init_node("qualisysSpoof")
        # zero zero of fake qualisys frame
        self.anchorPoint = rospy.get_param("~anchorPoint")
        rospy.Subscriber("dji_osdk_ros/gps_position", NavSatFix, self.gpsCallback, queue_size = 1)
        rospy.Subscriber("dji_osdk_ros/attitude", QuaternionStamped, self.rotCallback, queue_size = 1)
        # rospy.Subscriber("dji_osdk_ros/height_above_takeoff", Float32, self.heightCallback, queue_size = 1)

        # standing in for the feedback qualisys stream
        self.posePublisher = rospy.Publisher("feedback", Transform, queue_size = 1)
        self.quat = Quaternion()
        self.transform = Transform()
        self.startHeightSet = False
        self.startHeight = 0.0
        # self.height = 0.0

    def point_rotation(self, origin, pt, ang):
        # returns the pt rotated about the origin by ang (ang in degrees)
        c = math.cos(math.radians(ang))
        s = math.sin(math.radians(ang))
        # translate to origin
        pt_temp = [pt[0] - origin[0], pt[1] - origin[1]]
        pt_spun = [ pt_temp[0]*c - pt_temp[1]*s, pt_temp[0]*s + pt_temp[1]*c ]
        # translate back to frame
        pt_spun = [pt_spun[0] + origin[0], pt_spun[1] + origin[1]]
        return pt_spun

    def meters2lat_lon(self, meters_point):
        # returns point converted to lat lon from meters coordinate system
        transformer = Transformer.from_crs(3857, 4326)
        latlon_point = transformer.transform(meters_point[0], meters_point[1])
        return latlon_point

    def lat_lon2meters(self, latlon_point):
        # returns point converted to meters from lat lon coordinate system
        transformer = Transformer.from_crs(4326, 3857)
        meters_point = transformer.transform(latlon_point[0], latlon_point[1])
        return meters_point

    def rotCallback(self, data):
        self.quat = data.quaternion

    def gpsCallback(self, data):
        # convert and compare against home point
        apMeters = self.lat_lon2meters([self.anchorPoint[0], self.anchorPoint[1]])
        cpMeters = self.lat_lon2meters([data.latitude, data.longitude])
        currentOffset = [cpMeters[0] - apMeters[0], cpMeters[1] - apMeters[1]]

        # rotate current offset by a bit
        currentOffset = self.point_rotation([0,0], currentOffset, rospy.get_param("~angleOffset"));

        # save init height
        if (not self.startHeightSet):
            self.startHeightSet = True
            self.startHeight = data.altitude

        # performing dark magics
        self.transform.translation.x = currentOffset[1]
        self.transform.translation.y = currentOffset[0]
        self.transform.translation.z = data.altitude - self.startHeight
        self.transform.rotation = self.quat
        self.posePublisher.publish(self.transform)

if __name__ == '__main__':
    qualer = QualisysSpoof()
    rospy.spin()
