
/** @file mission_node.cpp
*  @version 4.0
*  @date June 2020
*
*  @brief node of hotpoint 1.0/waypoint 1.0.
*
*  @Copyright (c) 2020 DJI
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
*
*/
#include <math.h>       /* atan2 */
#include <dji_osdk_ros/bus_driver.h>
#include <dji_osdk_ros/FlightTaskControl.h>
#include <dji_osdk_ros/ObtainControlAuthority.h>
#include <dji_osdk_ros/Overwatch.h>
#include <geometry_msgs/Transform.h>
#include <geometry_msgs/Vector3.h>
#include <geometry_msgs/Quaternion.h>
#include <tf2/LinearMath/Quaternion.h>
#include <string>
#include <iostream>
#include <fstream>

#include <Eigen3/Eigen/Geometry>
#include <Eigen3/Eigen/Dense>

using namespace DJI::OSDK;
using namespace DJI::OSDK::Telemetry;

// global variables
ros::ServiceClient     waypoint_upload_client;
ros::ServiceClient     waypoint_action_client;
ros::ServiceClient     flight_control_client;
ros::ServiceClient     mission_waypoint_client;
ros::ServiceClient     obtain_ctrl_authority_client;
// need to take another look at synchronization...
// ros::Publisher         syncPub; // thing to publish to sync topic "positionSync"

sensor_msgs::NavSatFix gps_pos; // gps position from drone
geometry_msgs::Transform ex_pos; // position from external source (qualisys or some such)
geometry_msgs::Transform track_pos; // target position
geometry_msgs::Transform previous_ex_pos;
geometry_msgs::Transform previous_track_pos;

geometry_msgs::Transform pd_pos; // desired pose from pd relative to qual center (only x on rotation)

std::vector<float> expectedVel; // expected position after applying relative position move
float feedbackYaw; // yaw as measured by pos ref
float feedbackPitch; // pitch as measured by pos ref
float feedbackRoll; // roll as measured by pos ref
float trackingYaw; // yaw to track
float trackingPitch; // pitch to track
float trackingRoll; // roll to track
int calibrationWindow = 3; // the last _ position calls that are included in coordinate system calibration
std::vector<float> calibrationWindowValues;
float angleOffset;
float controlLoopTime;
float updateRateParam;
std::vector<std::vector<float>> poseRotationMatrix; // need to run calibration to get
std::vector<std::vector<std::vector<float>>> poseRotationMatrixValues; // need to run calibration to get
float poseMagnitudeScaler; // need to run calibration to get
float poseRotationRadians;
bool rotationMatrixCalibratedFlag = false;

bool                   gps_home_flag = false;
bool                   pos_home_flag = false;
bool                   localMissionFlag = false; // whether running position mission currently
bool                   gpsMissionFlag = false; // whether running gps mission currently
string                 missionTypeParam;
ros::Subscriber        gps_pos_subscriber;
ros::Subscriber        pos_feedback_subscriber;
ros::Subscriber        pos_tracking_subscriber;
ros::Subscriber        pd_position_subscriber;
ros::Subscriber        pd_heading_subscriber;

// temp subscribers to pick up puredata parameter information, will be removed once the right parameters are settled on
ros::Subscriber        pd_parameter_MinSpeed_subscriber;
ros::Subscriber        pd_parameter_MaxSpeed_subscriber;
ros::Subscriber        pd_parameter_MaxYawSpeed_subscriber;
ros::Subscriber        pd_parameter_SpeedScale_subscriber;
ros::Subscriber        pd_parameter_UpdateRate_subscriber;
ros::Subscriber        pd_parameter_Damping_subscriber;
ros::Subscriber        pd_parameter_Frequency_subscriber;
ros::Subscriber        pd_parameter_AngleOffset_subscriber;

int                    droneId; // stupid but will work for now

int                    waypointMissionUpdateCounter = 0; // counts iterations before we check waypoints again
int                    waypointMissionUpdateLimit = 10; // number of counts before we check ...
int                    positionMissionUpdateCounter = 0; // counts iterations before we check positions again
int                    positionMissionUpdateLimit = 10; // ...

std::vector<WayPointSettings> generatedWaypts; // waypoints to be generated, used to check if we're at waypoint or not
int                    waypointNumber;
int                    currentWaypoint = 0; // how many waypoints have we reached
long double            initialWaypointDistance; // distance to first waypoint
long double            initialPositionDistance; // distance to first position
double                 waypointSpeedParam;
double                 positionThresholdParam;
double                 angleOffsetParam;
double		             maxSpeedParam;
double                 speedScalarParam;
double                 maxYawSpeedParam;

// inverse controller variables
double                 dt;
double                 damping;
double                 frequency;
double                 kp;
double                 kd;
double                 ksg;
double                 kdg;
double                 g;


// Utility function for
// converting degrees to radians
long double toRadians(const long double degree)
{
  long double one_deg = (M_PI) / 180;
  return (one_deg * degree);
}

float normalizeAngle(float angle)
{
  float newAngle = std::fmod(angle, (2.0*M_PI));
  newAngle = std::fmod((newAngle + (2.0*M_PI)), (2.0*M_PI));
  if(newAngle >= M_PI)
  {
    newAngle -= (2.0*M_PI);
  }
  return newAngle;
}

float eucDistance(std::vector<float> a, std::vector<float> b)
{
  // do the math bb
  return pow(pow(a[0] - b[0],2) + pow(a[1] - b[1],2) + pow(a[2] - b[2],2),0.5);
}

long double gpsDistance(long double lat1, long double long1, long double lat2, long double long2)
{
  // Convert the latitudes
  lat1 = toRadians(lat1);
  long1 = toRadians(long1);
  lat2 = toRadians(lat2);
  long2 = toRadians(long2);

  // Haversine Formula
  long double dlong = long2 - long1;
  long double dlat = lat2 - lat1;

  long double ans = pow(sin(dlat / 2), 2) +
  cos(lat1) * cos(lat2) *
  pow(sin(dlong / 2), 2);

  ans = 2 * asin(sqrt(ans));

  long double R = 6371; // radius of earth

  // Calculate the result
  ans = ans * R;

  // recalculate in meters
  ans = ans * 1000;

  return ans;
}

std::vector<std::vector<float>> rotateAxisAngle(std::vector<float> u, float angleRad)
{
  const float sinA = sinf( angleRad );
  const float cosA = cosf( angleRad );
  const float oneMinusCosA = 1.0f - cosA;

  // rotationMatrix
  std::vector<std::vector<float>> rotationMatrix
  {
    {(u[0] * u[0] * oneMinusCosA) + cosA, (u[1] * u[0] * oneMinusCosA) - (sinA * u[2]), (u[2] * u[0] * oneMinusCosA) + (sinA * u[1])},
    {(u[0] * u[1] * oneMinusCosA) + (sinA * u[2]), (u[1] * u[1] * oneMinusCosA) + cosA, (u[2] * u[1] * oneMinusCosA) - (sinA * u[0])},
    {(u[0] * u[2] * oneMinusCosA) - (sinA * u[1]), (u[1] * u[2] * oneMinusCosA) + (sinA * u[0]), (u[2] * u[2] * oneMinusCosA) + cosA}
  };
  return rotationMatrix;
}

// dot product of two vector array.
float dotProduct(std::vector<float> vect_A, std::vector<float> vect_B)
{

    float product = 0;
    int vectorSize = vect_A.size(); // both vectors must be of same length
    for (int i = 0; i < vectorSize; i++)

        product = product + vect_A[i] * vect_B[i];
    return product;
}

float flatAngleBetween(std::vector<float> a, std::vector<float> b)
{
  // flatten each vector into the horizontal plane because im lazy
  std::vector<float> aFlat = std::vector<float> {a[0], a[1], 0.0};
  std::vector<float> bFlat = std::vector<float> {b[0], b[1], 0.0};
  float dotProd = dotProduct(aFlat, bFlat);
  return normalizeAngle(acosf(dotProd/(Magnitude3(a)*Magnitude3(b))));
}

// cross product of two vector array.
std::vector<float> crossProduct(std::vector<float> vect_A, std::vector<float> vect_B)
{
  std::vector<float> crossP
  {
    vect_A[1] * vect_B[2] - vect_A[2] * vect_B[1],
    vect_A[2] * vect_B[0] - vect_A[0] * vect_B[2],
    vect_A[0] * vect_B[1] - vect_A[1] * vect_B[0]
  };
  return crossP;
}

std::vector<float> normalizeVector(std::vector<float> vec)
{
  float mag = Magnitude3(vec);
  std::vector<float> result {
    vec[0]/mag,
    vec[1]/mag,
    vec[2]/mag
  };
  return result;
}

float Magnitude3(std::vector<float> v)
{
  return pow(pow(v[0], 2) + pow(v[1], 2) + pow(v[2], 2),0.5);
}

std::vector<float> MatrixMultiply(std::vector <std::vector<float>> a, std::vector<float> b)
{
    const int n = a.size();     // a cols
    const int m = a[0].size();  // a rows
    std::vector <float> c(n);
    // the PAIN
    for (auto k = 0; k < m; ++k) // row loop
    {
      for (auto i = 0; i < n; ++i) // col loop
      {
        // ROS_INFO("matrix (%d, %d): %f", k, i, a[i][k]);
        c[k] += a[i][k] * b[i];
      }
      // ROS_INFO("result val: %f", c[k]);
    }
    return c;
}

// std::vector<std::vector<float>> rotateAlign(std::vector<float> v1, std::vector<float> v2)
// {
//     std::vector<float> axis = normalizeVector(crossProduct(v1, v2));
//     float dotProd = dotProduct(normalizeVector(v1), normalizeVector(v2));
//     if (dotProd < -1.0)
//     {
//       dotProd = -1.0;
//     }
//     else if (dotProd > 1.0)
//     {
//       dotProd = 1.0;
//     }
//     poseRotationRadians = acosf(dotProd);
//     std::vector<std::vector<float>> result = rotateAxisAngle(axis, poseRotationRadians); // returns rotation matrix
//
//     ROS_INFO("axis of rotation: (%f, %f, %f)", axis[0], axis[1], axis[2]);
//     ROS_INFO("dot product: (%f)", dotProd);
//     ROS_INFO("angle rad: (%f)", poseRotationRadians);
//     return result;
// }

void pdPositionCallback(const geometry_msgs::Transform::ConstPtr& msg)
{
  pd_pos = *msg; // annnnnd thats it folks
}

// for those times where we actually fly outside lmao
void gpsPosCallback(const sensor_msgs::NavSatFix::ConstPtr& msg)
{
  gps_pos = *msg;

  if (!gpsMissionFlag)
  {
    return; // peace out if not running mission
  }

  if (!gps_home_flag)
  {
    int generatedWaypointNumber = static_cast<int>(generatedWaypts.size());
    if (generatedWaypointNumber > 0)
    {
      initialWaypointDistance = gpsDistance(gps_pos.latitude, gps_pos.longitude, generatedWaypts[0].latitude, generatedWaypts[0].longitude);
      ROS_INFO("current position: (%f, %f)", static_cast<float32_t>(gps_pos.latitude), static_cast<float32_t>(gps_pos.longitude));
      ROS_INFO("first waypoint position: (%f, %f)", static_cast<float32_t>(generatedWaypts[0].latitude), static_cast<float32_t>(generatedWaypts[0].longitude));

      ROS_INFO("distance to first waypoint: %f", static_cast<float32_t>(initialWaypointDistance));
      gps_home_flag = true;
    }
  }
  else
  {
    // do checking to see if we're near a waypoint, pause if so
    if(waypointMissionUpdateCounter >= waypointMissionUpdateLimit)
    {

      waypointMissionUpdateCounter = 0;
      int generatedWaypointNumber = static_cast<int>(generatedWaypts.size());
      if(generatedWaypointNumber > currentWaypoint) // if we actually have waypoints left
      {
        long double tolerance = 0.5; // reached waypoint detection threshold
        long double nextWaypointDistance;
        if (currentWaypoint == 0)
        {
          nextWaypointDistance = initialWaypointDistance;
        }
        else
        {
          nextWaypointDistance =
          gpsDistance(generatedWaypts[currentWaypoint-1].latitude, generatedWaypts[currentWaypoint-1].longitude, generatedWaypts[currentWaypoint].latitude, generatedWaypts[currentWaypoint].longitude);
        }

        long double progressDistance =
        gpsDistance(gps_pos.latitude, gps_pos.longitude, generatedWaypts[currentWaypoint].latitude, generatedWaypts[currentWaypoint].longitude);

        // now update sync topic with progress through waypoints
        std_msgs::Float32 syncUpdate;
        syncUpdate.data = static_cast<float32_t>(currentWaypoint) + static_cast<float32_t>(static_cast<float32_t>(1) - progressDistance/nextWaypointDistance);
        // syncPub.publish(syncUpdate);

        if(progressDistance < tolerance)
        {
          currentWaypoint++;
          ROS_INFO("reached new waypoint! have now reached: \t%d/%d\n", currentWaypoint, generatedWaypointNumber);
        }
      }
    }
    else
    {
      waypointMissionUpdateCounter++;
    }
  }
}

void posFeedbackCallback(const geometry_msgs::Transform::ConstPtr& msg)
{
  previous_ex_pos = ex_pos;
  ex_pos = *msg; // alllllways load the message

  // ROS_INFO("Current feedback postion %f, %f, %f", ex_pos.translation.x, ex_pos.translation.y, ex_pos.translation.z);
  //
  // // for fuckin symetry i guess
  // if (!localMissionFlag)
  // {
  //   return; // peace out if not running mission
  // }

  // FINE ill DO it MYself all ALONE
  feedbackYaw = atan2(2.0*(ex_pos.rotation.x*ex_pos.rotation.y + ex_pos.rotation.z*ex_pos.rotation.w), 1 - 2.0*(ex_pos.rotation.y*ex_pos.rotation.y + ex_pos.rotation.z*ex_pos.rotation.z));
  feedbackPitch = asin(2.0*(ex_pos.rotation.x*ex_pos.rotation.z - ex_pos.rotation.w*ex_pos.rotation.y));
  feedbackRoll = atan2(2.0*(ex_pos.rotation.x*ex_pos.rotation.w + ex_pos.rotation.y*ex_pos.rotation.z), 1 - 2.0*(ex_pos.rotation.z*ex_pos.rotation.z + ex_pos.rotation.w*ex_pos.rotation.w));
}

void posTrackingCallback(const geometry_msgs::Transform::ConstPtr& msg)
{
  previous_track_pos = track_pos;
  track_pos = *msg;

  // // for fuckin symetry i guess
  // if (!localMissionFlag)
  // {
  //   return; // peace out if not running mission
  // }

  trackingYaw = atan2(2.0*(track_pos.rotation.x*track_pos.rotation.y + track_pos.rotation.z*track_pos.rotation.w), 1 - 2.0*(track_pos.rotation.y*track_pos.rotation.y + track_pos.rotation.z*track_pos.rotation.z));
  trackingPitch = asin(2.0*(track_pos.rotation.x*track_pos.rotation.z - track_pos.rotation.w*track_pos.rotation.y));
  trackingRoll = atan2(2.0*(track_pos.rotation.x*track_pos.rotation.w + track_pos.rotation.y*track_pos.rotation.z), 1 - 2.0*(track_pos.rotation.z*track_pos.rotation.z + track_pos.rotation.w*track_pos.rotation.w));
}

bool setupWaypointMission(int responseTimeout)
{
  ros::spinOnce();

  while(!ros::param::has("/robot_list/robot_"+to_string(droneId)+"/numWaypoints"))
  {
    ROS_INFO("waiting for waypoints to load...");
  }

  // Waypoint Mission : Initialization
  dji_osdk_ros::MissionWaypointTask waypointTask;
  setWaypointInitDefaults(waypointTask);

  // Waypoint Mission: Create Waypoints
  ROS_INFO("Creating Waypoints..\n");
  generatedWaypts = loadWaypoints();
  // generatedWaypts = createWaypoints();

  // Waypoint Mission: Upload the waypoints
  ROS_INFO("Uploading Waypoints..\n");
  uploadWaypoints(generatedWaypts, responseTimeout, waypointTask);

  // Waypoint Mission: Init mission
  ROS_INFO("Initializing Waypoint Mission..\n");
  if (initWaypointMission(waypointTask).result)
  {
    ROS_INFO("Waypoint upload command sent successfully");
  }
  else
  {
    ROS_WARN("Failed sending waypoint upload command");
    return false;
  }

  return true;
}

// global waypoint mission

bool runWaypointMission()
{

  // Waypoint Mission: Start
  if (missionAction(DJI_MISSION_TYPE::WAYPOINT, MISSION_ACTION::START).result)
  {
    ROS_INFO("Mission start command sent successfully");
  }
  else
  {
    ROS_WARN("Failed sending mission start command");
    return false;
  }

  gpsMissionFlag = true;
  localMissionFlag = false;
  return true;
}

// relative position mission

bool runPositionMission()
{
  std::vector<float> velocityCommand {0.0, 0.0, 0.0};
  localMissionFlag = true;
  gpsMissionFlag = false;
  int currentPositionIndex = 0;
  int numPositions;
  ros::param::get("/robot_list/robot_" + to_string(droneId) + "/numPositions", numPositions);


  while (localMissionFlag && ros::ok())
  {
    ros::Rate updateRate(updateRateParam);
    // let topics update
    ros::spinOnce();

    dt = controlLoopTime;
    g = 1.0/(1.0 + kd*dt + kp*dt*dt);
    ksg = kp*g;
    kdg = (kd + kp*dt)*g;

    std::vector<float> currentPos
    {
      static_cast<float>(ex_pos.translation.x),
      static_cast<float>(ex_pos.translation.y),
      static_cast<float>(ex_pos.translation.z)
    };
    ROS_INFO("Current postion %f, %f, %f", currentPos[0], currentPos[1], currentPos[2]);

    std::vector<float> currentVel
    {
      static_cast<float>(ex_pos.translation.x - previous_ex_pos.translation.x),
      static_cast<float>(ex_pos.translation.y - previous_ex_pos.translation.y),
      static_cast<float>(ex_pos.translation.z - previous_ex_pos.translation.z)
    };

    std::vector<float> targetPos;
    std::vector<float> targetVel;
    float targetYaw;

    // wip, just align w external reference
        // check expected velocity verses actual velocity
        // if (Magnitude3(velocityCommand) != 0.0 && Magnitude3(currentVel) != 0.0)
        // {
        //   float ang = flatAngleBetween(velocityCommand, currentVel);
        //   updateAngleScaler(ang);
        //   ROS_INFO("new angle discrepancy: %f", angleOffset);
        // }

    // next depends on what mode we're in ------------------------------------
    if (missionTypeParam.compare("tracking") == 0)
    {
      targetPos = std::vector<float>
      {
        static_cast<float>(track_pos.translation.x),
        static_cast<float>(track_pos.translation.y),
        static_cast<float>(track_pos.translation.z)
      };

      targetVel = std::vector<float>
      {
        static_cast<float>(track_pos.translation.x - previous_track_pos.translation.x),
        static_cast<float>(track_pos.translation.y - previous_track_pos.translation.y),
        static_cast<float>(track_pos.translation.z - previous_track_pos.translation.z)
      };

      targetYaw = trackingYaw;
    }
    else if (missionTypeParam.compare("puredata") == 0)
    {
      // let pure data handle positioning, we just load from topic
      targetPos = std::vector<float>
      {
        static_cast<float>(pd_pos.translation.x),
        static_cast<float>(pd_pos.translation.y),
        static_cast<float>(pd_pos.translation.z)
      };
      targetYaw = pd_pos.rotation.x;
      targetVel = std::vector<float> {0.0, 0.0, 0.0};
    }
    else if (missionTypeParam.compare("points") == 0)
    {
      std::vector<float> tempPos;
      ros::param::get("/robot_list/robot_" + to_string(droneId) + "/positions/position_" + to_string(currentPositionIndex), tempPos);
      // pull out x,y,z
      targetPos = std::vector<float> {
        tempPos[1],
        tempPos[2],
        tempPos[3]
      };
      //----------------------------------
      targetVel = std::vector<float> {0.0, 0.0, 0.0};

      //----------------------------------
      targetYaw = normalizeAngle((M_PI/180.0)*tempPos[4]) + angleOffsetParam;
      // check if we're close to position
      float checkDist = eucDistance(currentPos, targetPos);
      if (checkDist <= positionThresholdParam) //&& abs(currentYaw - targetYaw) <= 0.1) // close ish
      {
        currentPositionIndex += 1; // go to next position
      }
      if (currentPositionIndex >= numPositions)
      {
        localMissionFlag = false;
      }
    }

    float currentYaw = feedbackYaw;
    ROS_INFO("Target position %f, %f, %f", targetPos[0], targetPos[1], targetPos[2]);
    ROS_INFO("Target velocity %f, %f, %f", targetVel[0], targetVel[1], targetVel[2]);
    ROS_INFO("Target yaw: %f", targetYaw);
    ROS_INFO("Current yaw: %f", currentYaw);
    ROS_INFO("Elapsed frame time: %f", static_cast<float>(dt));

    float yawCommand = normalizeAngle(currentYaw - targetYaw);
    ROS_INFO("yaw command: %f", yawCommand);

    std::vector<float> force
    {
      // using commanded velocity instead of actual
      static_cast<float>((targetPos[0] - currentPos[0])*ksg + (targetVel[0] - velocityCommand[0])*kdg),
      static_cast<float>((targetPos[1] - currentPos[1])*ksg + (targetVel[1] - velocityCommand[1])*kdg),
      static_cast<float>((targetPos[2] - currentPos[2])*ksg + (targetVel[2] - velocityCommand[2])*kdg)
    };
    // velocityCommand = MatrixMultiply(rotateAxisAngle(std::vector<float> {0.0, 0.0, 1.0}, angleOffset), velocityCommand);
    force = MatrixMultiply(rotateAxisAngle(std::vector<float> {0.0, 0.0, 1.0}, angleOffsetParam), force);
    velocityCommand = std::vector<float> {velocityCommand[0]*speedScalarParam + dt*force[0], velocityCommand[1]*speedScalarParam + dt*force[1], velocityCommand[2]*speedScalarParam + dt*force[2]};

    if (Magnitude3(velocityCommand) > maxSpeedParam)
    {
      velocityCommand = normalizeVector(velocityCommand);
      velocityCommand = std::vector<float> {maxSpeedParam*velocityCommand[0], maxSpeedParam*velocityCommand[1], maxSpeedParam*velocityCommand[2]};
    }
    ROS_INFO("calling vel command %f, %f, %f", velocityCommand[0], velocityCommand[1], velocityCommand[2]);
    relativeVelocity(velocityCommand[0], velocityCommand[1], velocityCommand[2], yawCommand, dt);
    updateRate.sleep();
  }
  return true;
}

void relativeVelocity(float x, float y, float z, float yaw, float dt)
{
  // convert relative position to velocity command
  std::vector<float> posVec {x, y, z};
  float totalSpeed = Magnitude3(posVec);
  float speed;
  // float dt = controlLoopTime*1000.0;
  if (totalSpeed > maxSpeedParam)
  {
    speed = maxSpeedParam;
  }
  else
  {
    speed = totalSpeed;
  }

  std::vector<float> velDir = normalizeVector(posVec);
  std::vector<float> velVec = {velDir[0]*speed, velDir[1]*speed, velDir[2]*speed};

  dji_osdk_ros::FlightTaskControl flightTaskControl;
  dji_osdk_ros::JoystickParams joystickParams;

  joystickParams.x = velVec[0];
  joystickParams.y = velVec[1];
  joystickParams.z = velVec[2];

  if (abs(yaw) > maxYawSpeedParam)
  {
    if (yaw < 0)
    {
      yaw = -maxYawSpeedParam;
    }
    else
    {
      yaw = maxYawSpeedParam;
    }
  }

  joystickParams.yaw = (180.0/M_PI)*yaw;

  ROS_INFO("rel vel: (yaw rate: %fdeg/s, speed: %fm/s, over %fsec, going %fm)", joystickParams.yaw, speed, dt, totalSpeed);
  ROS_INFO("---------------\n");

  flightTaskControl.request.task = dji_osdk_ros::FlightTaskControl::Request::TASK_VELOCITY_AND_YAWRATE_CONTROL;
  flightTaskControl.request.joystickCommand = joystickParams;
  flightTaskControl.request.velocityControlTimeMs = dt;
  flightTaskControl.request.posThresholdInM = 0.1;
  flightTaskControl.request.yawThresholdInDeg = 1;
  flight_control_client.call(flightTaskControl); // actually waits for drone to "reach" position

  // return flightTaskControl.response.result;
  // return std::vector<float> {(dt)*velVec[0], (dt)*velVec[1], (dt)*velVec[2]};
  return;
}

void relativePosition(float x, float y, float z, float yaw, float dt)
{
  // // convert relative position to velocity command
  std::vector<float> posVec {x, y, z};
  float totalDist = Magnitude3(posVec);
  // float speed;
  // // float dt = controlLoopTime*1000.0;
  // if ((totalDist/dt) > maxSpeedParam)
  // {
  //   speed = maxSpeedParam;
  // }
  // else
  // {
  //   speed = totalDist/dt;
  // }
  //
  // std::vector<float> velDir = normalizeVector(posVec);
  // std::vector<float> velVec = {velDir[0]*speed, velDir[1]*speed, velDir[2]*speed};
  //
  dji_osdk_ros::FlightTaskControl flightTaskControl;
  dji_osdk_ros::JoystickParams joystickParams;

  joystickParams.x = x;
  joystickParams.y = y;
  joystickParams.z = z;
  //
  // if (abs(yaw) > maxYawSpeedParam)
  // {
  //   if (yaw < 0)
  //   {
  //     yaw = -maxYawSpeedParam;
  //   }
  //   else
  //   {
  //     yaw = maxYawSpeedParam;
  //   }
  // }
  joystickParams.yaw = (180.0/M_PI)*yaw;


  ROS_INFO("rel pos: (yaw rate: %fdeg/s, speed: %fm/s, over %fsec, going %fm)", joystickParams.yaw, totalDist/dt, dt, totalDist);
  ROS_INFO("---------------\n");

  flightTaskControl.request.task = dji_osdk_ros::FlightTaskControl::Request::TASK_POSITION_AND_YAW_CONTROL;
  flightTaskControl.request.joystickCommand = joystickParams;
  flightTaskControl.request.velocityControlTimeMs = dt;
  flightTaskControl.request.posThresholdInM = 0.1;
  flightTaskControl.request.yawThresholdInDeg = 1;
  flight_control_client.call(flightTaskControl); // actually waits for drone to "reach" position

  // return flightTaskControl.response.result;
  // return std::vector<float> {(dt)*velVec[0], (dt)*velVec[1], (dt)*velVec[2]};
  return;
}

void setWaypointDefaults(WayPointSettings* wp)
{
  wp->damping         = 0.25;
  wp->yaw             = 0.0;
  wp->gimbalPitch     = 0;
  wp->turnMode        = 0;
  wp->hasAction       = 0;
  wp->actionTimeLimit = 100;
  wp->actionNumber    = 0;
  wp->actionRepeat    = 0;
  for (int i = 0; i < 16; ++i)
  {
    wp->commandList[i]      = 0;
    wp->commandParameter[i] = 0;
  }
}

void setWaypointInitDefaults(dji_osdk_ros::MissionWaypointTask& waypointTask)
{
  waypointTask.velocity_range     = waypointSpeedParam;
  waypointTask.idle_velocity      = waypointSpeedParam;
  waypointTask.action_on_finish   = dji_osdk_ros::MissionWaypointTask::FINISH_NO_ACTION;
  waypointTask.mission_exec_times = 1;
  waypointTask.yaw_mode		  = dji_osdk_ros::MissionWaypointTask::YAW_MODE_RC;
  waypointTask.trace_mode	  = dji_osdk_ros::MissionWaypointTask::TRACE_COORDINATED;
  waypointTask.action_on_rc_lost  = dji_osdk_ros::MissionWaypointTask::ACTION_AUTO;
  waypointTask.gimbal_pitch_mode  = dji_osdk_ros::MissionWaypointTask::GIMBAL_PITCH_FREE;
}

std::vector<DJI::OSDK::WayPointSettings> loadWaypoints()
{

  std::vector<double> wpt; // single waypoint
  // Let's create a vector to store our waypoints in.
  std::vector<DJI::OSDK::WayPointSettings> wp_list;
  int numPts;
  ros::param::get("/robot_list/robot_"+to_string(droneId)+"/numWaypoints", numPts);
  for (int i = 0; i < numPts; i++) {
    ros::param::get("/robot_list/robot_"+to_string(droneId)+"/waypoints/waypoint_" + to_string(i), wpt);
    // ROS_INFO("Waypoint read as (Ind, Lat, Lon, Alt, Yaw): %f %f %f %f %d\n ", wpt[0], wpt[1], wpt[2], wpt[3], wpt[4]);
    WayPointSettings wp;
    setWaypointDefaults(&wp);
    wp.index     = i;
    wp.latitude = wpt[1];
    wp.longitude = wpt[2];
    wp.altitude = wpt[3];
    wp.yaw = wpt[4];
    wp_list.push_back(wp);
  }
  return wp_list;

}

void uploadWaypoints(std::vector<DJI::OSDK::WayPointSettings>& wp_list, int responseTimeout, dji_osdk_ros::MissionWaypointTask& waypointTask)
{
  dji_osdk_ros::MissionWaypoint waypoint;
  double target_yaw = 0.0;
  // for (std::vector<WayPointSettings>::iterator wp = wp_list.begin();
  //      wp != wp_list.end(); ++wp)
  // {
  int numPts;
  DJI::OSDK::WayPointSettings wp;
  ros::param::get("/robot_list/robot_"+to_string(droneId)+"/numWaypoints", numPts);
  for (int i = 0; i < numPts; i++) {

    wp = wp_list[i];
    // target_yaw += 15.0;
    waypoint.latitude            = wp.latitude;
    waypoint.longitude           = wp.longitude;
    waypoint.altitude            = wp.altitude;
    waypoint.damping_distance    = wp.damping;
    waypoint.target_yaw          = wp.yaw;
    waypoint.target_gimbal_pitch = 0;
    waypoint.turn_mode           = wp.turnMode;
    waypoint.has_action          = 0;
    waypointTask.mission_waypoint.push_back(waypoint);
    ROS_INFO("Waypoint created at (Lat, Lon, Alt, Dmp, Yaw): %f %f %f %f %d\n ", waypoint.latitude, waypoint.longitude, waypoint.altitude, waypoint.damping_distance, waypoint.target_yaw);
  }
}

ServiceAck initWaypointMission(dji_osdk_ros::MissionWaypointTask& waypointTask)
{
  dji_osdk_ros::MissionWpUpload missionWpUpload;
  missionWpUpload.request.waypoint_task = waypointTask;
  waypoint_upload_client.call(missionWpUpload);
  if (!missionWpUpload.response.result)
  {
    ROS_WARN("ack.info: set = %i id = %i", missionWpUpload.response.cmd_set,
    missionWpUpload.response.cmd_id);
    ROS_WARN("ack.data: %i", missionWpUpload.response.ack_data);
  }
  return ServiceAck(
    missionWpUpload.response.result, missionWpUpload.response.cmd_set,
    missionWpUpload.response.cmd_id, missionWpUpload.response.ack_data);
}

ServiceAck missionAction(DJI::OSDK::DJI_MISSION_TYPE type,
  DJI::OSDK::MISSION_ACTION   action)
{
  dji_osdk_ros::MissionWpAction missionWpAction;
  dji_osdk_ros::MissionHpAction missionHpAction;
  missionWpAction.request.action = action;
  waypoint_action_client.call(missionWpAction);
  if (!missionWpAction.response.result)
  {
    ROS_WARN("ack.info: set = %i id = %i", missionWpAction.response.cmd_set,
    missionWpAction.response.cmd_id);
    ROS_WARN("ack.data: %i", missionWpAction.response.ack_data);
  }
  return { static_cast<bool>(missionWpAction.response.result),
    missionWpAction.response.cmd_set,
    missionWpAction.response.cmd_id,
    missionWpAction.response.ack_data };
}

bool takeoff()
{
  dji_osdk_ros::FlightTaskControl flightTaskControl;
  flightTaskControl.request.task = dji_osdk_ros::FlightTaskControl::Request::TASK_TAKEOFF;
  flight_control_client.call(flightTaskControl);

  return flightTaskControl.response.result;
}

bool land()
{
  dji_osdk_ros::FlightTaskControl flightTaskControl;
  flightTaskControl.request.task = dji_osdk_ros::FlightTaskControl::Request::TASK_FORCE_LANDING;
  flight_control_client.call(flightTaskControl);

  return flightTaskControl.response.result;
}

bool stopMotion()
{
  if (gpsMissionFlag)
  {
    dji_osdk_ros::MissionWpAction waypointAction;
    waypointAction.request.action = dji_osdk_ros::MissionWpAction::Request::ACTION_STOP; // stahhp
    waypoint_action_client.call(waypointAction);
    gpsMissionFlag = false;
  }

  if (localMissionFlag)
  {
    localMissionFlag = false;
  }
  return true;
}

bool goHome()
{
  dji_osdk_ros::FlightTaskControl flightTaskControl;
  flightTaskControl.request.task = dji_osdk_ros::FlightTaskControl::Request::TASK_GOHOME_AND_CONFIRM_LANDING;
  flight_control_client.call(flightTaskControl);

  return flightTaskControl.response.result;
}

void updateAngleScaler(float newValue)
{
  // remove the last value and insert new one
  calibrationWindowValues.pop_back();
  auto firstIndex = calibrationWindowValues.begin();
  calibrationWindowValues.insert(firstIndex, newValue);
  float sumer = 0.0;
  for (int i = 0; i < calibrationWindowValues.size(); i++)
  {
    sumer += calibrationWindowValues[i];
  }
  angleOffset = sumer/calibrationWindowValues.size();
  // angleOffset = calibrationWindowValues[firstIndex]; // just take most recent
}

// WIP
void calibratePositionReference()
{

  std::vector<float> deltaPosition = {1.0, 0.0, 0.0};
  std::vector<float> startPosition
  {
    static_cast<float>(ex_pos.translation.x),
    static_cast<float>(ex_pos.translation.y),
    static_cast<float>(ex_pos.translation.z)
  };
  // go forward at 1m/s for 1s
  relativeVelocity(deltaPosition[0], deltaPosition[1], deltaPosition[2], 0.0, 1.0);
  ros::Rate(1).sleep();
  ros::spinOnce(); // let topics update
  std::vector<float> deltaPositionActual
  {
    static_cast<float>(ex_pos.translation.x) - startPosition[0],
    static_cast<float>(ex_pos.translation.y) - startPosition[1],
    static_cast<float>(ex_pos.translation.z) - startPosition[2]
  };
  relativeVelocity(-deltaPosition[0], -deltaPosition[1], -deltaPosition[2], 0.0, 1.0);
  ros::Rate(1).sleep();
  ros::spinOnce();

  // recalculate the rotation offset
  // check expected velocity verses actual velocity
  ROS_INFO("calibrating angle offset");
  if (Magnitude3(deltaPosition) != 0.0 && Magnitude3(deltaPositionActual) != 0.0)
  {
    float ang = -flatAngleBetween(deltaPositionActual, deltaPosition);
    ROS_INFO("new angle offset: %f", ang);
    angleOffsetParam = static_cast<double>(ang);
    // updateAngleScaler(ang);
  }

}

void pdMaxSpeedCallback(const std_msgs::Float32::ConstPtr& msg)
{
  // update MaxSpeed parameters
  maxSpeedParam = msg->data;
  return;
}

void pdMaxYawSpeedCallback(const std_msgs::Float32::ConstPtr& msg)
{
  // update MaxYawSpeed parameters
  maxYawSpeedParam = msg->data;
  return;
}

void pdSpeedScaleCallback(const std_msgs::Float32::ConstPtr& msg)
{
  // update SpeedScale parameters
  speedScalarParam = msg->data;
  return;
}

void pdDampingCallback(const std_msgs::Float32::ConstPtr& msg)
{
  // update SpeedScale parameters
  if (msg->data < 0.0)
  {
    return;
  }
  damping = static_cast<double>(msg->data);
  kp = (6.0*frequency)*(6.0*frequency)* 0.25;
  kd = 4.5*frequency*damping;
  return;
}

void pdFrequencyCallback(const std_msgs::Float32::ConstPtr& msg)
{
  // update SpeedScale parameters
  if (msg->data < 0.0)
  {
    return;
  }
  frequency = static_cast<double>(msg->data);
  kp = (6.0*frequency)*(6.0*frequency)* 0.25;
  kd = 4.5*frequency*damping;
  return;
}

void pdAngleOffsetCallback(const std_msgs::Float32::ConstPtr& msg)
{
  // update SpeedScale parameters
  angleOffsetParam = msg->data;
  return;
}

void pdUpdateRateCallback(const std_msgs::Float32::ConstPtr& msg)
{
  // update UpdateRate parameters
  if (msg->data == 0.0)
  {
    return;
  }
  updateRateParam = msg->data;
  controlLoopTime = 1.0/updateRateParam;
  return;
}


bool overwatchFunction(dji_osdk_ros::Overwatch::Request &req, dji_osdk_ros::Overwatch::Response &res)
{

  // always request control?
  dji_osdk_ros::ObtainControlAuthority obtainCtrlAuthority;
  obtainCtrlAuthority.request.enable_obtain = true;
  obtain_ctrl_authority_client.call(obtainCtrlAuthority);

  // switch based on requested action
  switch (req.action) {
    case 1:
    {
      ROS_INFO("received overwatch: start waypoint");
      // start waypoint
      int responseTimeout = 1;
      setupWaypointMission(responseTimeout); // for now!
      runWaypointMission();
      break;
    }
    case 2:
    {
      ROS_INFO("received overwatch: start position");
      // start position
      runPositionMission();
      break;
    }
    case 3:
    {
      ROS_INFO("received overwatch: stop all motion");
      // stop all motion
      stopMotion();
      break;
    }
    case 4:
    {
      ROS_INFO("received overwatch: land");
      // land
      land();
      break;
    }
    case 5:
    {
      ROS_INFO("received overwatch: go home");
      // go home
      goHome();
      break;
    }
    case 6:
    {
      ROS_INFO("received overwatch: takeoff");
      // go home
      takeoff();
      break;
    }
    case 7:
    {
      ROS_INFO("received overwatch: calibrate pos ref");
      calibratePositionReference();
      break;
    }
    default:
      break;
  }

  res.result = true;
  return res.result;
}

int main(int argc, char** argv)
{

  ros::init(argc, argv, "mission_node");
  ros::NodeHandle nh;

  ros::param::get("base_speed", waypointSpeedParam);
  ros::param::get("max_speed", maxSpeedParam);
  ros::param::get("max_yaw_speed", maxYawSpeedParam);
  ros::param::get("angle_offset", angleOffsetParam);
  ros::param::get("speed_scalar", speedScalarParam);
  ros::param::get("drone_id", droneId);
  ros::param::get("mission_type", missionTypeParam);
  ros::param::get("position_threshold", positionThresholdParam);
  ros::param::get("loop_rate", updateRateParam);
  controlLoopTime = 1.0/updateRateParam;
  // additions
  ros::param::get("controller_damping", damping);
  ros::param::get("controller_frequency", frequency);
  kp = (6.0*frequency)*(6.0*frequency)* 0.25;
  kd = 4.5*frequency*damping;
  expectedVel = std::vector<float> {0.0, 0.0, 0.0};
  angleOffset = 0.0;
  calibrationWindowValues = std::vector<float> {};
  for (int i = 0; i < calibrationWindow; i++)
  {
    calibrationWindowValues.push_back(0.0);
  }

  // set puredata position to zero zero so it doesn't immediately fly away
  pd_pos.translation.x = 0.0;
  pd_pos.translation.y = 0.0;
  pd_pos.translation.z = 1.5;
  pd_pos.rotation.x = 0.0;

  // ROS stuff
  // obtain control authority?
  obtain_ctrl_authority_client = nh.serviceClient<dji_osdk_ros::ObtainControlAuthority>(
    "obtain_release_control_authority");
  waypoint_upload_client = nh.serviceClient<dji_osdk_ros::MissionWpUpload>(
    "dji_osdk_ros/mission_waypoint_upload");
  waypoint_action_client = nh.serviceClient<dji_osdk_ros::MissionWpAction>(
    "dji_osdk_ros/mission_waypoint_action");
  flight_control_client = nh.serviceClient<dji_osdk_ros::FlightTaskControl>(
    "flight_task_control");

  // syncPub = nh.advertise<std_msgs::Float32>("sync_signal", 1);
  gps_pos_subscriber = nh.subscribe<sensor_msgs::NavSatFix>("dji_osdk_ros/gps_position", 1, &gpsPosCallback);
  pos_feedback_subscriber = nh.subscribe<geometry_msgs::Transform>("feedback", 1, &posFeedbackCallback);
  pos_tracking_subscriber = nh.subscribe<geometry_msgs::Transform>("tracking", 1, &posTrackingCallback);
  pd_position_subscriber = nh.subscribe<geometry_msgs::Transform>("pd/target", 1, &pdPositionCallback);

  // pd_parameter_MinSpeed_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/minSpeed", 1, &pdMinSpeedCallback);
  pd_parameter_MaxSpeed_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/maxSpeed", 1, &pdMaxSpeedCallback);
  pd_parameter_MaxYawSpeed_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/maxYawSpeed", 1, &pdMaxYawSpeedCallback);
  pd_parameter_SpeedScale_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/speedScale", 1, &pdSpeedScaleCallback);
  pd_parameter_UpdateRate_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/updateRate", 1, &pdUpdateRateCallback);
  pd_parameter_AngleOffset_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/angleOffset", 1, &pdAngleOffsetCallback);
  pd_parameter_Damping_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/controllerDamping", 1, &pdDampingCallback);
  pd_parameter_Frequency_subscriber = nh.subscribe<std_msgs::Float32>("pd/parameters/controllerFrequency", 1, &pdFrequencyCallback);

  ros::ServiceServer overwatchService = nh.advertiseService("overwatch", overwatchFunction);

  // wait for services to be up
  obtain_ctrl_authority_client.waitForExistence();
  waypoint_upload_client.waitForExistence();
  waypoint_action_client.waitForExistence();
  flight_control_client.waitForExistence();
  // wait for waypoints to be up as well
  ROS_INFO("bus driver ready to go!");

  ros::spin();

  return 0;
}
